"""Viper Object Serializers
~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains various loaders that helps loading viper collection
objects (such as Hosts, Task etc.) from a file, and also dumpers that helps
dump tha same to a file.
"""

from __future__ import annotations
from collections import namedtuple
from enum import Enum
from json import dumps
from json import JSONDecodeError
from json import load
from json import loads

import typing as t

__all__ = ["Serializers"]

T = t.TypeVar("T")


def _load_json(jsonobj: t.Union[str, t.TextIO]) -> t.Any:
    if isinstance(jsonobj, str):

        return loads(jsonobj)

    return load(jsonobj)


def _dump_json(data: object, fileobj: t.Optional[t.TextIO] = None) -> str:

    json = dumps(data, indent=4)
    if fileobj is not None:
        fileobj.write(json)
        fileobj.flush()
    return json


def _json_or_literal(val: object) -> object:
    if isinstance(val, str):
        try:
            return loads(val)
        except (JSONDecodeError, TypeError):
            pass
    return val


def _load_csv(txtobj: t.Union[str, t.TextIO]) -> t.List[t.Dict[str, object]]:
    from csv import DictReader

    if isinstance(txtobj, str):
        from io import StringIO

        txtobj = StringIO(txtobj)

    rows = DictReader(txtobj)
    return [{k: _json_or_literal(v) for k, v in row.items()} for row in rows]


def _dump_csv(
    data: t.Sequence[t.Dict[str, object]], fileobj: t.Optional[t.TextIO] = None
) -> str:
    from csv import writer
    from io import StringIO

    out = StringIO()

    target = writer(out)
    target.writerow(data[0].keys())
    for dict_ in data:
        target.writerow(list(map(dumps, dict_.values())))
    if fileobj is not None:
        fileobj.write(out.getvalue())
        fileobj.flush()
    return out.getvalue()


def _load_yaml(yamlobj: t.Union[str, t.TextIO]) -> t.Any:
    from yaml import safe_load

    return safe_load(yamlobj)


def _dump_yaml(data: object, fileobj: t.Optional[t.TextIO] = None) -> str:
    from yaml import safe_dump

    yaml: str = safe_dump(data, default_flow_style=False)
    if fileobj is not None:
        fileobj.write(yaml)
        fileobj.flush()
    return yaml


Serializer = namedtuple("Serializer", ("load", "dump"))


class Serializers(Enum):
    """A list of serializers to help loading and dumping objects
    from or to a file.
    """

    json = Serializer(load=_load_json, dump=_dump_json)
    csv = Serializer(load=_load_csv, dump=_dump_csv)
    yml = Serializer(load=_load_yaml, dump=_dump_yaml)



__author__ = "Arijit Basu"
__description__ = "Viper is a handy tool for easily running infrastructure management tasks and commands."
__email__ = "sayanarijit@gmail.com"
__homepage__ = "https://github.com/sayanarijit/viper"
__license__ = "MIT"
__version__ = "v0.28.2"

from viper.collections import Host  # noqa: F401
from viper.collections import Hosts  # noqa: F401
from viper.collections import meta  # noqa: F401
from viper.collections import Result  # noqa: F401
from viper.collections import Results  # noqa: F401
from viper.collections import Runner  # noqa: F401
from viper.collections import Runners  # noqa: F401
from viper.collections import Task  # noqa: F401
from viper.collections import WhereConditions  # noqa: F401

__all__ = [
    "meta",
    "Host",
    "Hosts",
    "Result",
    "Results",
    "Runner",
    "Runners",
    "Task",
    "WhereConditions",
]
__doc__ = f"""::

     ▄   ▄█ █ ▄▄  ▄███▄   █▄▄▄▄   ▄█    ▄   ▄████  █▄▄▄▄ ██     ▄█▄    ████▄ █▀▄▀█ █▀▄▀█ ██      ▄   ██▄   ▄███▄   █▄▄▄▄
      █  ██ █   █ █▀   ▀  █  ▄▀   ██     █  █▀   ▀ █  ▄▀ █ █    █▀ ▀▄  █   █ █ █ █ █ █ █ █ █      █  █  █  █▀   ▀  █  ▄▀
 █     █ ██ █▀▀▀  ██▄▄    █▀▀▌    ██ ██   █ █▀▀    █▀▀▌  █▄▄█   █   ▀  █   █ █ ▄ █ █ ▄ █ █▄▄█ ██   █ █   █ ██▄▄    █▀▀▌
  █    █ ▐█ █     █▄   ▄▀ █  █    ▐█ █ █  █ █      █  █  █  █   █▄  ▄▀ ▀████ █   █ █   █ █  █ █ █  █ █  █  █▄   ▄▀ █  █
   █  █   ▐  █    ▀███▀     █      ▐ █  █ █  █       █      █   ▀███▀           █     █     █ █  █ █ ███▀  ▀███▀     █
    █▐        ▀            ▀         █   ██   ▀     ▀      █                   ▀     ▀     █  █   ██                ▀
    ▐                                                     ▀                               ▀


.. image:: https://img.shields.io/pypi/v/viper-infra-commander.svg
    :target: https://pypi.org/project/viper-infra-commander

.. image:: https://img.shields.io/pypi/pyversions/viper-infra-commander.svg
    :target: https://pypi.org/project/viper-infra-commander

.. image:: https://travis-ci.com/sayanarijit/viper.svg?branch=master
    :target: https://travis-ci.com/sayanarijit/viper

.. image:: https://codecov.io/gh/sayanarijit/viper/branch/master/graph/badge.svg
    :target: https://codecov.io/gh/sayanarijit/viper

.. image:: https://img.shields.io/badge/code%20style-black-000000.svg
    :target: https://github.com/python/black

.. image:: https://readthedocs.org/projects/viper-infrastructure-commander/badge/?version=latest
    :target: https://viper-infrastructure-commander.readthedocs.io


{__description__}


Getting Started
~~~~~~~~~~~~~~~

Installation
^^^^^^^^^^^^

.. code-block:: bash

    pip install -U viper-infra-commander

    # Or install with batteries included

    pip install -U "viper-infra-commander[batteries]"


Initialization
^^^^^^^^^^^^^^

.. code-block:: bash

    # (Optional) enable tab completion
    eval "$(viper autocomplete $(basename $SHELL))"


    # See the help menu
    viper -h


    # Initialize SQLite DB
    viper init -f


Viper In Action (Basic Mode)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Define a set of hosts in csv format (json and yml are also supported)

.. code-block:: bash

    cat > hosts.csv << EOF

.. code-block::

    ip,hostname,login_name,identity_file
    192.168.0.11,host11,root,/root/.ssh/id_rsa.pub
    192.168.0.12,host12,root,/root/.ssh/id_rsa.pub
    192.168.0.13,host13,root,/root/.ssh/id_rsa.pub
    192.168.0.14,host14,root,/root/.ssh/id_rsa.pub
    192.168.0.15,host15,root,/root/.ssh/id_rsa.pub

.. code-block:: bash

    EOF


Define a task

.. code-block:: bash

    cat > task.py << EOF

.. code-block:: python

    from viper import Task

    def ping_command(host):
        return "ping", "-c", "1", host.ip

    def ping():
        return Task(
            name="Ping once",
            command_factory=ping_command
        )

.. code-block:: bash

    EOF

Perform the following actions:

- Run the task on the set of hosts in parallel with 5 workers,
- filter only the results where the task failed,
- re-run the task on them,
- store the results in DB

.. code-block:: bash

    viper hosts:from-file hosts.csv \\
            | viper hosts:run-task task.ping --max-worker 5 \\
            | viper results:where returncode IS_NOT 0 \\
            | viper results:re-run --indent 4


See the stdout of the final results from DB

.. code-block:: bash

    viper results \\
            | viper results:final \\
            | viper results:format "{{host.hostname}}: {{stdout}}"


Export the results to a csv file

.. code-block:: bash

    viper results --final \\
            | viper results:to-file results.csv --indent 4


Define a job using the Python API (CLI and Python API are almost similar)

.. code-block:: bash

    cat > job.py << EOF

.. code-block:: python

    from viper import WhereConditions
    from task import ping

    def ping_and_export(hosts):
        return (
            hosts.task(ping())
            .run(max_workers=5)
            .final()
            .to_file("results.csv")
        )

.. code-block:: bash

    EOF


Run the job using CLI

.. code-block:: bash

    viper hosts:from-file hosts.csv \\
            | viper run job.ping_and_export \\
            | viper results:format "{{host.hostname}}: {{stdout}}"


Viperfile In Action (Advanced Mode)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Define a project in viperfile

.. code-block:: bash

    cat > viperfile.py << EOF

.. code-block:: python


    from viper import Hosts, Task
    from viper.project import Project, arg


    foo = Project(prefix="foo")


    @foo.hostgroup(args=[arg("-f", "--file", default="hosts.csv")])
    def allhosts(args):
        return Hosts.from_file(args.file)


    def remote_exec_command(host, command):
        return (
            "ssh",
            "-i",
            host.identity_file,
            "-l",
            host.login_name,
            "-p",
            str(host.port),
            "-o",
            "StrictHostKeyChecking=no",
            "-o",
            "PubkeyAuthentication=yes",
            host.ip,
            command,
        )


    @foo.job(
        args=[
            arg("command", help="command to execute"),
            arg("-w", "--workers", type=int, default=1),
        ]
    )
    def remote_exec(hosts, args):
        return (
            hosts.task(
                Task(
                    name="Remote execute command",
                    command_factory=remote_exec_command,
                    timeout=5,
                ),
                args.command,
            )
            .run(max_workers=args.workers)
            .final()
        )

.. code-block:: bash

    EOF


See the auto generated custom commands

.. code-block:: bash

    viper --help


Run the job

.. code-block:: bash

    viper @foo:allhosts \\
            | viper @foo:remote_exec "uname -a" --workers 5 \\
            | viper results:to-file results.csv \\
            | viper results:format "{{task.name}} [{{host.hostname}}]: {{returncode}}: {{stdout}}"


Further Readings
~~~~~~~~~~~~~~~~
**API Docs with Examples ☞** https://viper-infrastructure-commander.readthedocs.io


Contributing To Viper
~~~~~~~~~~~~~~~~~~~~~
**Contribution Guidelines ☞** https://github.com/sayanarijit/viper/blob/master/CONTRIBUTING.md
"""



"""Viper Command-line Interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The command-line interface strictly follows the Python interface.

Examples
^^^^^^^^

.. code-block:: bash

    viper hosts:from-file("hosts.csv") \\
            | viper hosts:task task.ping \\
            | viper runners:run --max-workers 50 \\
            | viper results:final \\
            | viper results:to-file results.csv

The above CLI example is equivalent to the following Python example:

.. code-block:: python

    from viper import Hosts
    import task

    (
        Hosts.from_file("hosts.csv")
        .task(task.ping())
        .run(max_workers=50)
        .final()
        .to_file("results.csv")
    )
"""

from argparse import ArgumentParser
from argparse import Namespace
from pydoc import locate
from types import FunctionType
from viper import __version__
from viper import Hosts
from viper import Results
from viper import Runners
from viper import Task
from viper.cli_base import SubCommand
from viper.collections import Collection as ViperCollection
from viper.collections import WhereConditions
from viper.const import Config
from viper.db import ViperDB

import os
import sqlite3
import sys
import traceback
import typing as t

__all__ = ["func", "run"]


def func(objpath: str) -> FunctionType:
    """Resolved Python function from given string."""

    funcobj = locate(objpath)
    if not func:
        raise ValueError(f"could not resolve {repr(objpath)}.")

    if not isinstance(funcobj, FunctionType):
        raise ValueError(f"{repr(objpath)} is not a valid function.")

    return funcobj


class AutocompleteCommand(SubCommand):
    """generate the auto completion script"""

    name = "autocomplete"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "shell", help="the target shell", choices=["bash", "tcsh", "zsh", "fish"]
        )

    def __call__(self, args: Namespace) -> int:
        import subprocess

        try:
            import argcomplete  # noqa F401
        except ImportError:
            raise RuntimeError(
                '"argcomplete" is missing! run `pip install -U "viper-infra-commander[autocomplete]"`'
            )

        _shell = "bash" if args.shell == "zsh" else args.shell
        script = subprocess.run(
            ["register-python-argcomplete", "viper", "-s", _shell],
            stdout=subprocess.PIPE,
            check=True,
        ).stdout.decode("latin1")

        if args.shell == "zsh":
            script = f"autoload bashcompinit\nbashcompinit\n{script}"

        print(script)
        return 0


class InitCommand(SubCommand):
    """initialize the current workspace"""

    name = "init"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="remove or overwrite existing data",
        )

    def __call__(self, args: Namespace) -> int:
        try:
            ViperDB.init(ViperDB.url, force=args.force)
        except sqlite3.OperationalError:
            raise RuntimeError(
                "database already exists!"
                " use '-f'/'--force' to force re-create the database."
            )
        return 0


class LetsCommand(SubCommand):
    """perform any defined action"""

    name = "lets"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("action", type=func, help="action definition location")
        parser.add_argument("args", nargs="*", help="optional arguments for the action")

    def __call__(self, args: Namespace) -> int:
        res = args.action(*args.args)
        if res is not None:
            print("\n".join(map(str, res)) if isinstance(res, tuple) else res)
        return 0


class RunJobCommand(SubCommand):
    """[Hosts -> Results] run a job on the given hosts"""

    name = "run-job"
    aliases = ("run",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("job", type=func, help="job definition location")
        parser.add_argument("args", nargs="*", help="arguments to be passed to the job")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        results = args.job(Hosts.from_json(input()), *args.args)
        if not isinstance(results, Results):
            raise ValueError(
                f"a job must return {Results} object but got {type(results)}"
            )
        print(results.to_json(indent=args.indent))
        return 0


class TaskFromFuncCommand(SubCommand):
    """[Task] get the task from a Python function location"""

    name = "task:from-func"
    aliases = ("task",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "func", type=Task.from_func, help="load task from a Python function"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(args.func.to_json(indent=args.indent))
        return 0


class TaskResultsCommand(SubCommand):
    """[Task -> Results] get the past results of given task"""

    name = "task:results"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Task.from_json(input()).results().to_json(indent=args.indent))
        return 0


class TaskFormatCommand(SubCommand):
    """[Task -> str] format the data using the given template"""

    name = "task:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")

    def __call__(self, args: Namespace) -> int:
        print(Task.from_json(input()).format(args.template))
        return 0


class HostsFromFuncCommand(SubCommand):
    """[Hosts] get a group of hosts from a Python function location"""

    name = "hosts:from-func"
    aliases = ("hosts",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "func", type=Hosts.from_func, help="load hosts from a Python function"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(args.func.to_json(indent=args.indent))
        return 0


class HostsFromFileCommand(SubCommand):
    """[Hosts] get a group of hosts from a file"""

    name = "hosts:from-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_file(args.filepath).to_json(indent=args.indent))
        return 0


class HostsToFileCommand(SubCommand):
    """[Hosts -> Hosts] export the group of hosts to a file"""

    name = "hosts:to-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input()).to_file(args.filepath).to_json(indent=args.indent)
        )
        return 0


class HostsTaskCommand(SubCommand):
    """[Hosts -> Runners] assign a task to each host"""

    name = "hosts:task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "task", type=Task.from_func, help="load task from Python function"
        )
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the command factory"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:

        print(
            Hosts.from_json(input())
            .task(args.task, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class HostsRunTaskCommand(SubCommand):
    """[Hosts -> Results] assign a task to each host and run"""

    name = "hosts:run-task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "task", type=Task.from_func, help="load task from a Python function"
        )
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the command factory"
        )
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .run_task(args.task, *args.args, max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class HostsFilterCommand(SubCommand):
    """[Hosts -> Hosts] filter hosts by a given function"""

    name = "hosts:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class HostsCountCommand(SubCommand):
    """[Hosts -> int] count the number of hosts"""

    name = "hosts:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).count())
        return 0


class HostsOrderByCommand(SubCommand):
    """[Hosts -> Hosts] sort the hosts by the given properties"""

    name = "hosts:order-by"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "properties", nargs="*", help="items will be sorted in order"
        )
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .order_by(*args.properties, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class HostsSortCommand(SubCommand):
    """[Hosts -> Hosts] sort the hosts by the given function"""

    name = "hosts:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func, help="the sorting function")
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .sort(key=args.key, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class HostsPipeCommand(SubCommand):
    """[Hosts -> ?] pipe the hosts to the given handler"""

    name = "hosts:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj: t.Any = Hosts.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class HostsFormatCommand(SubCommand):
    """[Hosts -> str] format the data using the given template"""

    name = "hosts:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).format(args.template, sep=args.sep))
        return 0


class HostsWhereCommand(SubCommand):
    """[Hosts -> Hosts] select hosts matching the given query"""

    name = "hosts:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name/path of each item")
        parser.add_argument(
            "condition", choices=[o.value for o in WhereConditions],
        )
        parser.add_argument("values", nargs="*")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Hosts.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class HostsHeadCommand(SubCommand):
    """[Hosts -> Hosts] get the first 'n' hosts"""

    name = "hosts:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class HostsTailCommand(SubCommand):
    """[Hosts -> Hosts] get the last 'n' hosts"""

    name = "hosts:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class HostsResultsCommand(SubCommand):
    """[Hosts -> Results] get the past results of the hosts"""

    name = "hosts:results"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Hosts.from_json(input()).results().to_json(indent=args.indent))
        return 0


class RunnersFromFileCommand(SubCommand):
    """[Runners] get a group of runners from a file"""

    name = "runners:from-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_file(args.filepath).to_json(indent=args.indent))
        return 0


class RunnersToFileCommand(SubCommand):
    """[Runners -> Runners] export the group of runners to a file"""

    name = "runners:to-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .to_file(args.filepath)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersFilterCommand(SubCommand):
    """[Runners -> Runners] filter runners by a given function"""

    name = "runners:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersCountCommand(SubCommand):
    """[Runners -> int] count the number of runners"""

    name = "runners:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).count())
        return 0


class RunnersOrderByCommand(SubCommand):
    """[Runners -> Runners] sort the runners by the given properties"""

    name = "runners:order-by"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "properties", nargs="*", help="items will be sorted in order"
        )
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .order_by(*args.properties, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersSortCommand(SubCommand):
    """[Runners -> Runners] sort the runners by the given funtion"""

    name = "runners:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func)
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .sort(key=args.key, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersPipeCommand(SubCommand):
    """[Runners -> ?] pipe the runners to the given handler"""

    name = "runners:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj: t.Any = Runners.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class RunnersFormatCommand(SubCommand):
    """[Runners -> str] format the data using the given template"""

    name = "runners:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).format(args.template, sep=args.sep))
        return 0


class RunnersWhereCommand(SubCommand):
    """[Runners -> Runners] select runners matching the given query"""

    name = "runners:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name/path of each item")
        parser.add_argument(
            "condition", choices=[o.value for o in WhereConditions],
        )
        parser.add_argument("values", nargs="*")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersHeadCommand(SubCommand):
    """[Runners -> Runners] get the first 'n' runners"""

    name = "runners:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class RunnersTailCommand(SubCommand):
    """[Runners -> Runners] get the last 'n' runners"""

    name = "runners:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class RunnersRunCommand(SubCommand):
    """[Runners -> Results] run the assigned tasks"""

    name = "runners:run"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Runners.from_json(input())
            .run(max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class RunnersHostsCommand(SubCommand):
    """[Runners -> Hosts] get the hosts from the runners"""

    name = "runners:hosts"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Runners.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class ResultsFromHistoryCommand(SubCommand):
    """[Results] get the past results from database"""

    name = "results:from-history"
    aliases = ("results",)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--final",
            action="store_true",
            help="get the final results only (shortcut to `viper results | viper results:final`)",
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_history(final=args.final).to_json(indent=args.indent))
        return 0


class ResultsFromFileCommand(SubCommand):
    """[Results] get a group of results from a file"""

    name = "results:from-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_file(args.filepath).to_json(indent=args.indent))
        return 0


class ResultsToFileCommand(SubCommand):
    """[Results -> Results] export the group of results to a file"""

    name = "results:to-file"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filepath")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .to_file(args.filepath)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsFilterCommand(SubCommand):
    """[Results -> Results] filter results by a given handler"""

    name = "results:filter"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("filter", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .filter(args.filter, *args.args)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsCountCommand(SubCommand):
    """[Results -> int] count the number of results"""

    name = "results:count"

    def add_arguments(self, parser: ArgumentParser) -> None:
        pass

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).count())
        return 0


class ResultsOrderByCommand(SubCommand):
    """[Results -> Results] sort the results by the given properties"""

    name = "results:order-by"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "properties", nargs="*", help="items will be sorted in order"
        )
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .order_by(*args.properties, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsSortCommand(SubCommand):
    """[Results -> Results] sort the results by the given function"""

    name = "results:sort"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--key", type=func)
        parser.add_argument(
            "--reverse", action="store_true", help="reverse the order after sort"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .sort(key=args.key, reverse=args.reverse)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsPipeCommand(SubCommand):
    """[Results -> ?] pipe the results to the given handler"""

    name = "results:pipe"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("handler", type=func)
        parser.add_argument(
            "args", nargs="*", help="arguments to be passed to the filter"
        )
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        obj: t.Any = Results.from_json(input()).pipe(args.handler, *args.args)

        if obj is None:
            return 0

        if isinstance(obj, ViperCollection):
            print(obj.to_json(indent=args.indent))
            return 0

        print(obj)
        return 0


class ResultsFormatCommand(SubCommand):
    """[Results -> str] format the data using the given template"""

    name = "results:format"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("template", help="use Python's string format syntax")
        parser.add_argument(
            "-s", "--sep", help="separator used to join the strings", default="\n"
        )

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).format(args.template, sep=args.sep))
        return 0


class ResultsWhereCommand(SubCommand):
    """[Results -> Results] select results matching the given query"""

    name = "results:where"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("key", help="property name of each item")
        parser.add_argument(
            "condition",
            choices=[o.value for o in WhereConditions],
            help="condition to apply",
        )
        parser.add_argument("values", nargs="*", help="values for the key")
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .where(args.key, WhereConditions(args.condition), args.values)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsHeadCommand(SubCommand):
    """[Results -> Results] get the first 'n' results"""

    name = "results:head"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).head(args.n).to_json(indent=args.indent))
        return 0


class ResultsTailCommand(SubCommand):
    """[Results -> Results] get the last 'n' results"""

    name = "results:tail"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-n", type=int, help="number of hosts", default=10)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).tail(args.n).to_json(indent=args.indent))
        return 0


class ResultsHostsCommand(SubCommand):
    """[Results -> Hosts] get the hosts from the results"""

    name = "results:hosts"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).hosts().to_json(indent=args.indent))
        return 0


class ResultsRunnersCommand(SubCommand):
    """[Results -> Runners] recreate the runners from the results"""

    name = "results:runners"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).runners().to_json(indent=args.indent))
        return 0


class ResultsReRunCommand(SubCommand):
    """[Results -> Results] recreate the runners from the results and run again"""

    name = "results:re-run"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--max-workers", type=int, default=Config.max_workers.value)
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(
            Results.from_json(input())
            .runners()
            .run(max_workers=args.max_workers)
            .to_json(indent=args.indent)
        )
        return 0


class ResultsByTaskCommand(SubCommand):
    """[Task -> Results] get the past results of given task"""

    name = "results:by-task"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.by_task(Task.from_json(input())).to_json(indent=args.indent))
        return 0


class ResultsFinal(SubCommand):
    """[Results -> Results] get the final results only (ignoring the previous retries)"""

    name = "results:final"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("-i", "--indent", type=int, default=None)

    def __call__(self, args: Namespace) -> int:
        print(Results.from_json(input()).final().to_json(indent=args.indent))
        return 0


def run() -> int:
    parser = ArgumentParser(
        "viper", description=f"Viper Infrastructure Commander {__version__}"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--debug",
        action="store_true",
        help="show traceback information when an exception is raised",
    )
    subparsers = parser.add_subparsers()

    # Autocomplete command
    AutocompleteCommand.attach_to(subparsers)

    # Init command
    InitCommand.attach_to(subparsers)

    # Init command
    LetsCommand.attach_to(subparsers)

    # Job run command
    RunJobCommand.attach_to(subparsers)

    # Task commands
    TaskFromFuncCommand.attach_to(subparsers)
    TaskResultsCommand.attach_to(subparsers)
    TaskFormatCommand.attach_to(subparsers)

    # Hosts commands
    HostsFromFuncCommand.attach_to(subparsers)

    HostsFromFileCommand.attach_to(subparsers)
    HostsToFileCommand.attach_to(subparsers)
    HostsWhereCommand.attach_to(subparsers)
    HostsFilterCommand.attach_to(subparsers)
    HostsOrderByCommand.attach_to(subparsers)
    HostsSortCommand.attach_to(subparsers)
    HostsCountCommand.attach_to(subparsers)
    HostsPipeCommand.attach_to(subparsers)
    HostsFormatCommand.attach_to(subparsers)
    HostsHeadCommand.attach_to(subparsers)
    HostsTailCommand.attach_to(subparsers)

    HostsTaskCommand.attach_to(subparsers)
    HostsRunTaskCommand.attach_to(subparsers)
    HostsResultsCommand.attach_to(subparsers)

    # Task runners commands
    RunnersFromFileCommand.attach_to(subparsers)
    RunnersToFileCommand.attach_to(subparsers)
    RunnersWhereCommand.attach_to(subparsers)
    RunnersFilterCommand.attach_to(subparsers)
    RunnersOrderByCommand.attach_to(subparsers)
    RunnersSortCommand.attach_to(subparsers)
    RunnersCountCommand.attach_to(subparsers)
    RunnersPipeCommand.attach_to(subparsers)
    RunnersFormatCommand.attach_to(subparsers)
    RunnersHeadCommand.attach_to(subparsers)
    RunnersTailCommand.attach_to(subparsers)

    RunnersRunCommand.attach_to(subparsers)
    RunnersHostsCommand.attach_to(subparsers)

    # Task results commands
    ResultsFromHistoryCommand.attach_to(subparsers)

    ResultsFromFileCommand.attach_to(subparsers)
    ResultsToFileCommand.attach_to(subparsers)
    ResultsWhereCommand.attach_to(subparsers)
    ResultsFilterCommand.attach_to(subparsers)
    ResultsOrderByCommand.attach_to(subparsers)
    ResultsSortCommand.attach_to(subparsers)
    ResultsCountCommand.attach_to(subparsers)
    ResultsPipeCommand.attach_to(subparsers)
    ResultsFormatCommand.attach_to(subparsers)
    ResultsHeadCommand.attach_to(subparsers)
    ResultsTailCommand.attach_to(subparsers)

    ResultsFinal.attach_to(subparsers)
    ResultsHostsCommand.attach_to(subparsers)
    ResultsRunnersCommand.attach_to(subparsers)
    ResultsReRunCommand.attach_to(subparsers)
    ResultsByTaskCommand.attach_to(subparsers)

    # Add the modules path
    sys.path.insert(0, os.path.realpath(Config.modules_path.value))

    # Import project specific commands
    if os.path.exists("viperfile.py"):
        from viper import project

        sys.path.insert(0, os.path.realpath("."))
        import viperfile

        for k, v in vars(viperfile).items():
            if not isinstance(v, project.Project):
                continue

            for subcommand in v.all_commands():
                subcommand.attach_to(subparsers)

    try:
        # Add support for auto completion
        import argcomplete

        argcomplete.autocomplete(parser)
    except ImportError:
        pass

    args = parser.parse_args()

    if not hasattr(args, "_handler"):
        parser.print_usage()
        return 1

    try:
        return int(args._handler(args))
    except Exception as e:
        print(f"error: {e}", file=sys.stderr)
        if args.debug:
            print(traceback.format_exc(), file=sys.stderr)
        return 1



from __future__ import annotations

import typing as t

T = t.TypeVar("T")


def optional(
    dict_: t.Dict[object, object],
    /,
    key: object,
    expect: t.Union[type, t.Tuple[type, ...]],
    *,
    parser: t.Optional[t.Callable[[t.Dict[object, object], object], object]] = None,
) -> t.Optional[T]:
    """Get a value safely from the given dict using the given key.
    If the value is not found or is None, return it anyway.

    :raises: ValueError
    """

    if parser is None:
        value = dict_.get(key)
    else:
        value = parser(dict_, key)

    if value is not None and not isinstance(value, expect):
        raise ValueError(
            f"{dict_}: {repr(key)}: {repr(value)}: invalid type {type(value)}, expecting {expect}"
        )
    return t.cast(T, value)


def required(
    dict_: t.Dict[object, object],
    /,
    key: object,
    expect: t.Union[type, t.Tuple[type, ...]],
    *,
    parser: t.Optional[t.Callable[[t.Dict[object, object], object], object]] = None,
    default: t.Optional[T] = None,
    default_factory: t.Optional[t.Callable[[], T]] = None,
) -> T:
    """Get a value safely from the given dict using the given key.
    If the value is not found or is None, raise value error.

    :raises: ValueError
    """
    value: t.Optional[T] = optional(dict_, key, expect, parser=parser)
    if value is None:
        if default is not None:
            return default
        if default_factory is not None:
            return default_factory()
        raise ValueError(f"{dict_}: {repr(key)}: value is required")
    return value


def flatten_dict(d: t.Dict[t.Any, t.Any]) -> t.Dict[str, object]:
    def items() -> t.Iterable[t.Tuple[str, object]]:
        for key, value in d.items():
            if not isinstance(key, str):
                raise ValueError(f"{key}: expected {str}, but got {type(key)}")
            if isinstance(value, dict):
                for subkey, subvalue in flatten_dict(value).items():
                    if not isinstance(subkey, str):
                        raise ValueError(
                            f"{subkey}: expected {str}, but got {type(subkey)}"
                        )
                    yield key + ":" + subkey, subvalue
            else:
                yield key, value

    return dict(items())


def unflatten_dict(d: t.Dict[object, object]) -> t.Dict[object, object]:
    dict_: t.Dict[object, object] = {}
    for k, v in d.items():
        if not isinstance(k, str):
            raise ValueError(f"{k}: expected {str}, but got {type(k)}")

        if ":" in k:
            _dict: t.Union[object, t.Dict[object, object]] = dict_
            fields = k.split(":")
            count = len(fields)
            for i in range(count):
                field = fields[i]
                if not isinstance(_dict, dict):
                    continue
                if field not in _dict:
                    _dict[field] = v if i == (count - 1) else {}
                _dict = _dict[field]
        else:
            dict_[k] = v
    return dict_



"""Viper Project APIs (the ``viperfile.py``)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Viper provides APIs for customizing the viper CLI based on the project
requirement. When viper detects a ``viperfile.py`` file in the current
working directory, it scans that file and adds the project specific
commands and options to it's own list.

We can see the commands by running ``viper --help``. These commands are
prefixed with ``@`` so that they are easily recognizable.


After defining this in ``viperfile.py`` we can use commands like ``viper @myproj:allhosts``
(example) to get the list of hosts which can be piped to other commands that recieves
a list of hosts from `stdin`.
"""


from __future__ import annotations
from argparse import ArgumentParser
from argparse import Namespace
from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from viper.cli_base import SubCommand
from viper.collections import Collection as ViperCollection
from viper.collections import Hosts
from viper.collections import Results

import typing as t

T = t.TypeVar("T")
C = t.TypeVar("C")
ArgType = t.Tuple[t.Tuple[str, ...], t.Dict[str, t.Any]]
HostsFuncType = t.Callable[[Namespace], Hosts]
HandlerFuncType = t.Callable[[T, Namespace], C]
JobFuncType = t.Callable[[Hosts, Namespace], Results]
ActionFuncType = t.Callable[[Namespace], T]


def arg(*args: str, **kwargs: t.Any) -> ArgType:
    """Argumenst to be passed to argparse"""
    return args, kwargs


@dataclass
class Project:
    """A project (namespace) for the viper sub commands

    :param str prefix: Prefix for the related sub commands.

    When we define a project, we basically define a namespace (a prefix)
    for the commands.
    """

    prefix: str
    action_commands: t.List[t.Type[SubCommand]] = field(default_factory=lambda: [])
    hostgroup_commands: t.List[t.Type[SubCommand]] = field(default_factory=lambda: [])
    filter_commands: t.List[t.Type[SubCommand]] = field(default_factory=lambda: [])
    handler_commands: t.List[t.Type[SubCommand]] = field(default_factory=lambda: [])
    job_commands: t.List[t.Type[SubCommand]] = field(default_factory=lambda: [])

    def all_commands(self) -> t.List[t.Type[SubCommand]]:
        """Return all sub commands"""
        return (
            self.action_commands
            + self.hostgroup_commands
            + self.filter_commands
            + self.handler_commands
            + self.job_commands
        )

    def hostgroup(
        self, args: t.Optional[t.Sequence[ArgType]] = None
    ) -> t.Callable[[HostsFuncType], HostsFuncType]:
        """Use this decorator to define host groups

        :param list args (optional): Arguments to be parsed by py:class:`argparse.ArgumentParser`
        """

        def wrapper(func: HostsFuncType) -> HostsFuncType:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class HostGroupCommand(SubCommand):
                __doc__ = f"[Hosts] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    print(func(args).to_json(indent=args.indent))
                    return 0

            self.hostgroup_commands.append(HostGroupCommand)
            return func

        return wrapper

    def handler(
        self,
        fromtype: type,
        totype: type,
        args: t.Optional[t.Sequence[ArgType]] = None,
    ) -> t.Callable[[HandlerFuncType[T, C]], HandlerFuncType[T, C]]:
        """Use this decorator to define handlers

        :param type fromtype: The type of object this handler is expecting.
        :param type totype: The type of object this handler returns.
        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.
        """

        if not issubclass(fromtype, ViperCollection):
            raise ValueError(f"{fromtype} does not have pipe option")

        def wrapper(func: HandlerFuncType[T, C]) -> HandlerFuncType[T, C]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class HandlerCommand(SubCommand):
                __doc__ = f"[{fromtype.__name__} -> {totype.__name__}] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    if not issubclass(fromtype, ViperCollection):
                        raise ValueError(f"{fromtype}: invalid fromtype")

                    obj: C = fromtype.from_json(input()).pipe(
                        lambda obj: func(obj, args)
                    )
                    if not isinstance(obj, totype):
                        raise ValueError(
                            f"invalid totype, expected {totype} but got {type(obj)}"
                        )

                    if isinstance(obj, ViperCollection):
                        print(obj.to_json(indent=args.indent))
                    else:
                        print(obj)

                    return 0

            self.handler_commands.append(HandlerCommand)
            return func

        return wrapper

    def job(
        self, args: t.Optional[t.Sequence[ArgType]] = None,
    ) -> t.Callable[[JobFuncType], JobFuncType]:
        """Use this decorator to define a job.

        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.
        """

        def wrapper(func: JobFuncType) -> JobFuncType:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class JobCommand(SubCommand):
                __doc__ = f"[Hosts -> Results] {doc}"
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])
                    parser.add_argument("-i", "--indent", type=int, default=None)

                def __call__(self, args: Namespace) -> int:
                    results = func(Hosts.from_json(input()), args)
                    if not isinstance(results, Results):
                        raise ValueError(
                            f"a job must return {Results} object but got {type(results)}"
                        )
                    print(results.to_json(indent=args.indent))
                    return 0

            self.job_commands.append(JobCommand)
            return func

        return wrapper

    def action(
        self,
        args: t.Optional[t.Sequence[ArgType]] = None,
        totype: t.Optional[type] = None,
    ) -> t.Callable[[ActionFuncType[T]], ActionFuncType[T]]:
        """Use this decorator to define an action.

        :param list args (optional): List of arguments for :py:class:`argparse.ArgumentParser`.
        """

        def wrapper(func: ActionFuncType[T]) -> ActionFuncType[T]:

            doc = func.__doc__.splitlines()[0] if func.__doc__ else ""

            class ActionCommand(SubCommand):
                __doc__ = f"[{totype.__name__}] {doc}" if totype else doc
                name = f"@{self.prefix}:{func.__name__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg[0], **arg[1])

                def __call__(self, args: Namespace) -> int:
                    res = func(args)
                    if res is None:
                        if totype is None:
                            return 0
                        else:
                            raise ValueError(f"expected an object of type {totype}")

                    if totype is None:
                        raise ValueError(
                            f"not expecting any result, but got a result of type {type(res)}"
                        )

                    if not isinstance(res, totype):
                        raise ValueError(
                            f"expected result of type {totype} but got result of type {type(res)}"
                        )

                    if isinstance(res, ViperCollection):
                        print(res.to_json())
                    elif isinstance(res, Iterable):
                        print("\n".join(map(str, res)))
                    else:
                        print(res)
                    return 0

            self.action_commands.append(ActionCommand)
            return func

        return wrapper
